from sqlalchemy.orm import registry
from sqlalchemy import Column, Integer, String, DateTime, SmallInteger, Float

mapper_registry = registry()
Base = mapper_registry.generate_base()


class RecordingSession(Base):
    __tablename__ = 'session'

    id = Column(Integer, primary_key=True)
    session_hash = Column(String(40), nullable=False)
    host = Column(String, nullable=False)
    port = Column(Integer, nullable=False)
    start = Column(DateTime, nullable=False)
    stop = Column(DateTime)

    def __repr__(self):
        return (f'RecordingSession(id={self.id!r}, '
                f'session_hash={self.session_hash!r}, host={self.host!r}, '
                f'port={self.port!r}, start={self.start!r}, '
                f'stop={self.stop!r})')


class Ident(Base):
    __tablename__ = 'ident'

    id = Column(Integer, primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    icao = Column(String(6), nullable=False)
    callsign = Column(String(8), nullable=False)
    typecode = Column(SmallInteger, nullable=False)
    category = Column(SmallInteger, nullable=False)

    def __repr__(self):
        return (f'Ident(id={self.id!r}, timestamp={self.id!r}, '
                f'icao={self.icao!r}, callsign={self.callsign!r}, '
                f'typecode={self.typecode!r}, category={self.category!r})')


class Velocity(Base):
    __tablename__ = 'velocity'

    id = Column(Integer, primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    icao = Column(String(6), nullable=False)
    speed = Column(SmallInteger)
    speed_type = Column(String)  #TODO determine possible values
    vertical_speed = Column(SmallInteger)
    vertical_speed_src = Column(String)  #TODO make enum
    angle = Column(Float)
    angle_src = Column(String)  #TODO make enum

    def __repr__(self):
        return (f'Velocity(id={self.id!r}, timestamp={self.timestamp!r}, '
                f'icao={self.icao!r}, speed={self.speed!r}, '
                f'speed_type={self.speed_type!r}, '
                f'vertical_speed={self.vertical_speed!r}, '
                f'angle={self.angle!r}, '
                f'vertical_speed_src={self.vertical_speed_src!r}, '
                f'angle_src={self.angle_src!r})')


class Position(Base):
    __tablename__ = 'position'

    id = Column(Integer, primary_key=True)
    timestamp = Column(DateTime, nullable=False)
    icao = Column(String(6), nullable=False)
    latitude = Column(Float)
    longitude = Column(Float)
    altitude = Column(Integer)  #TODO determine if small int works
    altitude_src = Column(String)  #TODO make enum

    def __repr__(self):
        return (f'Position(id={self.id!r}, timestamp={self.timestamp!r}, '
                f'icao={self.icao!r}, latitude={self.latitude!r}, '
                f'longitude={self.longitude!r}, altitude={self.altitude!r}, '
                f'altitude_src={self.altitude_src!r})')



from datetime import datetime as dt
from adsb_track.const import *
from adsb_track.database import DBSQL
from adsb_track.aircraft import Airspace


def recreate_airspace_from_messages(messages):
    """Recreates airspace from messages

    Args:
        messages (list of tuples): The messages to use

    Returns:
        adsb_track.Airspace: The airspace from the reconstructed messages.
    """
    airspace = Airspace()
    for msg in messages:
        type_ = msg[0]
        msg = msg[1:]
        ts = msg[DBSQL.TIMESTAMP_INDEX]
        icao = msg[DBSQL.ICAO_INDEX]
        if type_ == IDENT:
            callsign = msg[DBSQL.IDENT_INDICES[CALLSIGN]]
            airspace.update_callsign(icao, ts, callsign)
        elif type_ == VELOCITY:
            heading, speed, vs = [
                msg[DBSQL.VELOCITY_INDICES[x]]
                for x in (ANGLE, SPEED, VERTICAL_SPEED)
            ]
            airspace.update_velocity(icao, ts, heading, speed, vs)
        elif type_ == POSITION:
            lat, lon, alt = [
                msg[DBSQL.POSITION_INDICES[x]]
                for x in (LATITUDE, LONGITUDE, ALTITUDE)
            ]
            airspace.update_position(icao, ts, lat, lon, alt)
    return airspace


def recreate_airspace_from_session(db, session_uuid):
    """Recreates airspace from session messages

    Args:
        db (str): Path to database
        session_uuid (str): Session to replay

    Returns:
        adsb_track.Airspace: The airspace from the reconstructed messages.
    """
    return recreate_airspace_from_messages(db.replay_session(session_uuid))


def recreate_airspace_from_timestamp(db, datetime=None, previous=60):
    """Recreates airspace around a given time.

    Args:
        db (str): Path to database
        datetime (datetime.datetime or float): Ending time
        previous (int): The number of seconds to look back.

    Returns:
        adsb_track.Airspace: The airspace from the reconstructed messages.
    """
    stop = datetime if isinstance(datetime, (int, float)) else None
    if datetime is None:
        stop = dt.now().timestamp()
    elif isinstance(datetime, str):
        if datetime.lower() == 'last':
            stop = db.last_message()
        else:
            stop = dt.fromisoformat(datetime).timestamp()
    start = stop - previous
    return recreate_airspace_from_messages(db.replay_messages(start, stop))


if __name__ == '__main__':
    from adsb_track.database import DBSQLite
    db = DBSQLite('test.sqlite3')
    ac = recreate_airspace(db, 'last')


import argparse
from adsb_track.stream import FlightRecorder

supported_rawtypes = 'raw', 'beast', 'skysense'

parser = argparse.ArgumentParser(description='Capture and record ADS-B data.')
parser.add_argument('database',
                    help='The SQLite database file to record data.',
                    metavar='DATABASE')
parser.add_argument('--host',
                    default='localhost',
                    help='The host of the ADS-B source')
parser.add_argument('--port',
                    default=30002,
                    type=int,
                    help='The port of the ADS-B source')
parser.add_argument('--rawtype',
                    default='raw',
                    choices=supported_rawtypes,
                    help='The ADS-B output data format')
parser.add_argument('--latlon',
                    nargs=2,
                    type=float,
                    required=True,
                    help='Receiver latitude and longitude',
                    metavar=('LAT', 'LON'))

args = parser.parse_args()
if args.rawtype not in supported_rawtypes:
    raise ValueError(
        f"Rawtype {args.rawtype} not one of {', '.join(supported_rawtypes)}")
if abs(args.latlon[0]) > 90:
    raise ValueError(f"Provided latitude {args.latlon[0]} is outside -90 to 90")
if abs(args.latlon[1]) > 180:
    raise ValueError(
        f"Provided longitude {args.latlon[1]} is outside -180 to 180")

flights = FlightRecorder(args.host, args.database, args.latlon[0],
                         args.latlon[1], args.port, args.rawtype)

flights.record()



TIMESTAMP = 'timestamp'
ICAO = 'icao'

LAST_UPDATE = 'last_update'
MSG_TYPE = 'msg_type'

SESSION = 'session'
SESSION_HASH = 'session_hash'
HOST = 'host'
PORT = 'port'
START = 'start'
STOP = 'stop'
DURATION = 'duration'

IDENT = 'ident'
CALLSIGN = 'callsign'
TYPECODE = 'typecode'
CATEGORY = 'category'

VELOCITY = 'velocity'
SPEED = 'speed'
SPEED_TYPE = 'speed_type'
VERTICAL_SPEED = 'vertical_speed'
VERTICAL_SPEED_SRC = 'vertical_speed_src'
ANGLE = 'angle'
ANGLE_SRC = 'angle_src'

POSITION = 'position'
LATITUDE = 'latitude'
LONGITUDE = 'longitude'
ALTITUDE = 'altitude'
ALTITUDE_SRC = 'altitude_src'


from numpy import positive
import adsb_track.const as const

import pandas as pd
from sqlalchemy import create_engine, select, between
from sqlalchemy.orm import Session

from adsb_track.const import *
from adsb_track.database.schema import (Base, RecordingSession, Ident, Velocity,
                                        Position)


class Database:

    def record_session_start(self, session_hash, host, port, start):
        """Records the session socket and start time

        Args:
            session_hash (str): Session SHA-1 hash
            host (str): Session host
            port (int): Session port
            start (datetime.datetime): Session start time
        """
        self.session.add(
            RecordingSession(session_hash=session_hash,
                             host=host,
                             port=port,
                             start=start))

    def record_session_stop(self, session_hash, stop):
        """Records the session end time

        Args:
            session_hash (str): Active session SHA-1 hash
            stop (datetime.datetime): Active session end time
        """
        self.session.execute(
            select(RecordingSession).filter_by(
                session_hash=session_hash)).scalar_one().stop = stop

    def record_ident(self, ts, icao, callsign, tc, cat):
        """Records an identification message

        Args:
            ts (datetime.datetime): Message timestamp
            icao (str): Aircraft ICAO24 code
            callsign (str): Aircraft callsign
            tc (int): Aircraft typecode
            cat (int): Aircraft category
        """
        self.session.add(
            Ident(timestamp=ts,
                  icao=icao,
                  callsign=callsign,
                  typecode=tc,
                  category=cat))

    # Order meant to match pyModeS return
    def record_velocity(self, ts, icao, spd, angle, vs, spd_type, angle_src,
                        vs_src):
        """Records a velocity message

        Args:
            ts (datetime.datetime): Message timestamp
            icao (str): Aircraft ICAO24 code
            spd (int): Aircraft speed
            angle (float): Aircraft heading
            vs (int): Aircraft vertical speed
            spd_type (str): Type of speed recorded
            angle_src (str): Source of heading measurement
            vs_src (str): Source of vertical speed measurement
        """
        self.session.add(
            Velocity(timestamp=ts,
                     icao=icao,
                     speed=spd,
                     speed_type=spd_type,
                     vertical_speed=vs,
                     vertical_speed_src=vs_src,
                     angle=angle,
                     angle_src=angle_src))

    def record_position(self, ts, icao, lat, lon, alt, alt_src):
        """Records a position message
        
        Args:
            ts (datetime.datetime): Message timestamp
            icao (str): Aircraft ICAO24 code
            lat (float): Aircraft latitude
            lon (float): Aircraft longitude
            alt (int): Aircraft altitude
            alt_src (str): Source of altitude measurement
        """
        self.session.add(
            Position(timestamp=ts,
                     icao=icao,
                     latitude=lat,
                     longitude=lon,
                     altitude=alt,
                     altitude_src=alt_src))

    def replay_messages(self, start, stop):
        """Replays the message in a given time duration

        Args:
            start (datetime.datetime): Start time, seconds since UNIX epoch
            stop (datetime.datetime): Stop time, seconds since UNIX epoch
        
        Returns:
            pandas.DataFrame: The messages captured in the time duration
        """
        return tuple([
            pd.read_sql_query(
                select(x).where(between(x.timestamp, start, stop)),
                self.engine,
                index_col='id',
            ) for x in (Ident, Velocity, Position)
        ])

    def list_sessions(self):
        """Lists the recording sessions of the receiver
        
        Returns:
            pandas.DataFrame: The sessions found in the database.
        """

        df = pd.read_sql_table(RecordingSession.__tablename__,
                               self.engine,
                               index_col='id')
        string_col = [const.SESSION_HASH, const.HOST]
        df.loc[:, string_col] = df.loc[:, string_col].convert_dtypes()
        df[const.DURATION] = df[const.STOP] - df[const.START]
        return df

    def replay_session(self, session_hash):
        """Replays the messages of a specified session.
        
        Args:
            session_hash (str): Session SHA-1 hash
        
        Returns:
            pandas.DataFrame: the messages in the given session.
        """

        hash_sql_prefix = RecordingSession.session_hash.ilike(
            f'{session_hash}%')

        all_sessions = [
            x[0] for x in self.session.execute(
                select(RecordingSession.session_hash).where(hash_sql_prefix))
        ]

        if not all_sessions:  # No matches found
            raise ValueError('No matching sessions hashes found')
        if len(all_sessions) > 1:
            raise ValueError(
                f'Ambiguous SHA-1 hash prefix. Found sessions: {all_sessions}')

        timestamps = [
            x for x in self.session.execute(
                select(RecordingSession.start, RecordingSession.stop).where(
                    hash_sql_prefix))
        ][0]

        return self.replay_messages(timestamps.start, timestamps.stop)

    def close_session(self):
        self.session.commit()
        self.session.close()

    def __init__(self, dialect, url):
        if dialect == 'sqlite':
            database_url = f'sqlite:///{url}'
        self.engine = create_engine(database_url)
        Base.metadata.create_all(self.engine)
        self.session = Session(self.engine)



from adsb_track.const import *
def outliers(series):
    q1, q3 = [series.quantile(x, 'linear') for x in [0.25, 0.75]]
    three_halves_iqr = 1.5 * (q3 - q1)
    return (max(0, q1-three_halves_iqr),
            min(q3+three_halves_iqr, series.max()))

def remove_outliers(series):
    low, high = outliers(series)
    return series[series]

def average_position(df_track):
    return tuple([df_track[x].mean() for x in (LATITUDE, LONGITUDE)])


from datetime import datetime as dt
import hashlib
import struct

import pyModeS as pms
from pyModeS.extra.tcpclient import TcpClient

from adsb_track.database import Database


class FlightRecorder(TcpClient):

    self.TC_POS = tuple(range(9, 19)) + tuple(range(20, 23))
    self.TC_IDENT = tuple(range(1, 5))

    @staticmethod
    def create_session_hash(host: str, port: int, rawtype: str, start: float):
        """Creates a unique has for the session.

        Args:
            host (str): Session host
            port (int): Session port
            rawtype (str): Session rawtype
            start (float): Session start time.
        Returns:
            str: SHA-1 hash of the session.
        """
        h = hashlib.sha1()
        h.update(str.encode(host))
        h.update(int.to_bytes(port, 2, 'big'))
        h.update(str.encode(rawtype))
        h.update(bytes(struct.pack('d', start)))
        return h.hexdigest()

    def __init__(self,
                 host,
                 db,
                 gs_lat,
                 gs_lon,
                 port=30002,
                 rawtype='raw',
                 buffer=25):
        super(FlightRecorder, self).__init__(host, port, rawtype)
        now = dt.now()
        self.session_hash = self.create_session_hash(host, port, rawtype,
                                                     now.timestamp())
        self.gs_lat = gs_lat
        self.gs_lon = gs_lon
        self.db = Database('sqlite', db)
        self.db.record_session_start(self.session_hash, host, port, now)

    def process_msg(self, msg, ts, icao, tc):
        if tc in self.TC_POS:
            self.process_position(msg, ts, icao, tc)
        elif tc == 19:
            self.process_velocity(msg, ts, icao)
        elif tc in self.TC_IDENT:
            self.process_ident(msg, ts, icao, tc)

    def process_position(self, msg, ts, icao, tc):
        alt_src = 'BARO' if tc < 19 else 'GNSS'
        alt = pms.adsb.altitude(msg)
        lat, lon = pms.adsb.position_with_ref(msg, self.gs_lat, self.gs_lon)

        self.db.record_position(ts, icao, lat, lon, alt, alt_src)

    def process_velocity(self, msg, ts, icao):
        velocity = pms.adsb.velocity(msg, True)

        self.db.record_velocity(ts, icao, *velocity)

    def process_ident(self, msg, ts, icao, tc):
        callsign = pms.adsb.callsign(msg).strip('_')
        category = pms.adsb.category(msg)

        self.db.record_ident(ts, icao, callsign, tc, category)

    def handle_messages(self, messages):
        for msg, ts in messages:
            if len(msg) == 28 and pms.df(msg) == 17 and pms.crc(msg) == 0:
                icao = pms.adsb.icao(msg)
                tc = pms.adsb.typecode(msg)
                self.process_msg(msg, dt.fromtimestamp(ts), icao, tc)

    def record(self):
        try:
            self.run()
        except KeyboardInterrupt:
            self.db.record_session_stop(self.session_hash, dt.now())
            self.db.close_session()



from datetime import datetime as dt

import pandas as pd
import numpy as np

from adsb_track.const import *
import adsb_track.const as const


class SessionData:

    def __init__(self, df_ident, df_velocity, df_position):
        self.df_ident = df_ident
        self.df_velocity = df_velocity
        self.df_position = df_position
        self.unique_icao = np.unique(
            np.concatenate([
                x[const.ICAO].unique()
                for x in (df_ident, df_velocity, df_position)
            ]))

    def unique_icao(self):
        """Unique ICAO addresses in the message dataframes.

        Args:
            message_dataframes (Iterable[pandas.DataFrame]): The dataframes from
                the session output. The column with ICAO24 codes is expected to
                be 'icao'.

        Returns:
            Iterable[str]: The unique ICAO24 codes found across the input
                dataframes.
        """
        return self.unique_icao

    def isolate_icao(self, icao):
        """Isolates the messages of an aircraft from many dataframes.

        Args:
            icao (str): The ICAO24 code to isolate.

        Returns:
            Iterable[pandas.DataFrame]: A deepcopy of a subset of the input
                message_dataframes where the ICAO24 code matches the icao input
                parameter.
        """
        if icao not in self.unique_icao:
            return ValueError(f'The ICAO24 code {icao} could not be found in '
                              'any messages.')
        return tuple([
            x[x[const.ICAO] == icao].copy()
            for x in (self.df_ident, self.df_velocity, self.df_position)
        ])

    def build_track(self, icao):
        """Constructs the track of an aircraft from different types of messages.

        Args:
            icao (str): The ICAO24 code to isolate.

        Returns:
            pandas.DataFrame: A single dataframe constructed with the most
                recent information of each type.
        """
        if icao not in self.unique_icao:
            return ValueError(f'The ICAO24 code {icao} could not be found in '
                              'any messages.')
        df_ident, df_velocity, df_position = self.isolate_icao(icao)
        df_ident[const.MSG_TYPE] = const.IDENT
        df_velocity[const.MSG_TYPE] = const.VELOCITY
        df_position[const.MSG_TYPE] = const.POSITION
        df = pd.concat([df_ident, df_velocity, df_position])
        df.sort_values(const.TIMESTAMP, inplace=True)
        df.ffill(inplace=True)
        return df


class Aircraft:
    """A representation of an aircraft over time

    Args:
        icao (str): The ICAO24 code of the aircraft, used as its unique
            identifier.
    """

    def __init__(self, icao):
        self.icao = icao

        self.callsign_update = None
        self.callsign = None
        self.callsign_history = []

        self.position_update = None
        self.latitude = None
        self.longitude = None
        self.altitude = None
        self.position_history = []

        self.velocity_update = None
        self.heading = None
        self.velocity = None
        self.vertical_speed = None
        self.velocity_history = []

    def __str__(self):

        def title():
            icao_lowercase = self.icao.lower()
            if self.callsign is None:
                return self.icao.lower()
            return f'{self.callsign} / {icao_lowercase}'

        return (
            f"------ {title()} ------\n"
            f'  ({self.latitude}, {self.longitude})  {self.altitude} ft\n'
            f'  {self.heading} degrees, {self.velocity} knots, {self.vertical_speed} ft/sec'
        )

    def last_update(self):
        """The last time the aircraft reported.

        Returns:
            pandas.Timestamp: The latest timestamp of either the callsign,
                position, or velocity.
        """
        update_canidates = [
            x for x in (self.callsign_update, self.position_update,
                        self.velocity_update) if x is not None
        ]
        if len(update_canidates) > 0:
            return max(update_canidates)

    def to_json(self):
        """Exports the aircraft data in JSON format.

        Returns:
            dict: The aircraft data in JSON format
        """
        return {
            ICAO: self.icao,
            LAST_UPDATE: self.last_update().timestamp(),
            CALLSIGN: self.callsign,
            LATITUDE: self.latitude,
            LONGITUDE: self.longitude,
            ALTITUDE: self.altitude,
            ANGLE: self.heading,
            SPEED: self.velocity,
            VERTICAL_SPEED: self.vertical_speed,
        }

    def process_timestamp(ts):
        if isinstance(ts, pd._libs.tslibs.timestamps.Timestamp):
            return ts
        elif isinstance(ts, float):
            return pd.to_datetime(ts, unit='s')
        elif isinstance(ts, dt):
            return pd.to_datetime(ts)

    def is_update(ts, comparison):
        return (comparison is None) or (ts > comparison)

    def get_callsign_history(self):
        """Produces a time series of recorded call sign.

        Returns:
            pandas.DataFrame: A dataframe with timestamps and callsigns.
        """
        if self.callsign_history:
            return pd.DataFrame(self.callsign_history,
                                columns=[TIMESTAMP, CALLSIGN]).convert_dtypes()

    def get_position_history(self):
        """Produces a time series of position data.

        Returns:
            pandas.DataFrame: A dataframe with timestamps, latitude, longitude,
                and altitude.
        """
        if self.position_history:
            return pd.DataFrame(
                self.position_history,
                columns=[TIMESTAMP, LATITUDE, LONGITUDE, ALTITUDE])

    def get_velocity_history(self):
        """Produces a time series of velocity.

        Returns:
            pandas.DataFrame: A dataframe with heading, speed, and vertical
                speed.
        """
        if self.velocity_history:
            return pd.DataFrame(
                self.velocity_history,
                columns=[TIMESTAMP, ANGLE, VELOCITY, VERTICAL_SPEED])

    def get_track(self):
        """Provides a single dataframe with all of callsign, position, and
            altitude history.

        Returns:
            pandas.DataFrame: A dataframe of callsign, position, and altitude
                concatenated.
        """
        df = pd.concat(
            (self.get_callsign_history(), self.get_position_history(),
             self.get_velocity_history()))
        df.sort_values(TIMESTAMP, inplace=True)
        df.reset_index(drop=True, inplace=True)
        return df.convert_dtypes(convert_floating=False)

    def update_callsign(self, ts, callsign):
        """Updates the aircraft callsign at a given time.

        Args:
            ts (Pandas Timestamp or float or Python datetime): The time at which
                the callsign record occurred. If the input is a float, that is
                the seconds since the UNIX epoch.
            callsign (str): The aircraft callsign
        """
        ts = Aircraft.process_timestamp(ts)
        if Aircraft.is_update(ts, self.callsign_update):
            self.callsign_update = ts
            self.callsign = callsign
            self.callsign_history.append((ts, callsign))

    def update_position(self, ts, lat, lon, alt):
        """Updates the aircraft velocity at a given time.

        Args:
            ts (Pandas Timestamp, float, Python datetime): The time at which the
                velocity was recorded.
            lat (float): Latitude
            lon (float): Longitude
            alt (int): Altitude
        """
        ts = Aircraft.process_timestamp(ts)
        if Aircraft.is_update(ts, self.position_update):
            self.position_update = ts
            self.latitude = lat
            self.longitude = lon
            self.altitude = alt
            self.position_history.append((ts, lat, lon, alt))

    def update_velocity(self, ts, heading, velocity, vertical_speed):
        """Updates the aircraft position at a given time.

        Args:
            ts (Pandas Timestamp, float, Python datetime): The time at which the
                velocity was recorded.
            heading (float): Heading
            velocity (int): Velocity
            vertical_speed (int): Vertical speed
        """
        ts = Aircraft.process_timestamp(ts)
        if Aircraft.is_update(ts, self.velocity_update):
            self.velocity_update = ts
            self.heading = heading
            self.velocity = velocity
            self.vertical_speed = vertical_speed
            self.velocity_history.append(
                (ts, heading, velocity, vertical_speed))


class Airspace:
    """A representation of the entire airspace."""

    def __init__(self):
        self.flights = {}

    def __len__(self):
        return len(self.flights)

    def to_json(self):
        """JSON representation of all aircraft in airspace.

        Returns:
            list of dict: A JSON list of all the aircraft included in airspace.
        """
        return [x.to_json() for x in self.flights.values()]

    def aircraft_present(self):
        """Lists all aircraft in the airspace.

        Returns:
            dict_keys: A set of ICAO24 codes
        """
        return self.flights.keys()

    def get_aircraft(self, icao):
        """Attempts to find the specified aircraft

        Args:
            icao (str): ICAO24 address

        Returns:
            adsb_track.Aircraft or None: The aircraft if found
        """
        return self.flights.get(icao.upper())

    def check_aircraft(self, icao):
        """Finds aircraft in airspace or creates a new one.

        Returns:
            adsb_track.Aircraft: The aircraft (possibly a new one) in the
                airspace with the corresponding ICAO24 code.
        """
        icao_uppper = icao.upper()
        if icao_uppper not in self.flights:
            self.flights[icao_uppper] = Aircraft(icao_uppper)
        return self.flights[icao_uppper]

    def update_callsign(self, icao, ts, callsign):
        """Updates the callsign information of the aircraft.

        Args:
            icao (str): ICAO24 code
            ts (Pandas Timestamp or float or Python datetime): The time at which
                the callsign record occurred. If the input is a float, that is
                the seconds since the UNIX epoch.
            callsign (str): The aircraft callsign
        """
        self.check_aircraft(icao).update_callsign(ts, callsign)

    def update_position(self, icao, ts, lat, lon, alt):
        """Updates the position information of an aircraft.

        Args:
            icao (str): ICAO24 code
            ts (Pandas Timestamp, float, Python datetime): The time at which the
                velocity was recorded.
            lat (float): Latitude
            lon (float): Longitude
            alt (int): Altitude
        """
        self.check_aircraft(icao).update_position(ts, lat, lon, alt)

    def update_velocity(self, icao, ts, heading, velocity, vertical_speed):
        """Updates the velocity information of an aircraft

        Args:
            icao (str): ICAO24 code
            ts (Pandas Timestamp, float, Python datetime): The time at which the
                velocity was recorded.
            heading (float): Heading
            velocity (int): Velocity
            vertical_speed (int): Vertical speed
        """
        self.check_aircraft(icao).update_velocity(ts, heading, velocity,
                                                  vertical_speed)

    def __str__(self):
        return ('\n' * 2).join([str(x) for x in self.flights.values()])
